### Python Code Style Guide for chatbot_ner

**Why follow a style guide ?**

- To keep the coding style consistent across the whole repo
- Make it more readable and maintainable.
- Ensure that others can work with the code you write
- Avoid opinionated agruments about the cosmetic quality of code and documentation in code reviews

Since the project has been running for quite a long time without any code analysis checks, there are inconsistencies
everywhere. And since we can't fix everything at once, moving forward we would want to fix such issues part by part
as and when we make changes to said parts.

**General Instructions:**

- Read the guidelines and externals links presented in this document.
- Keep revisiting them from time to time.
- Update this document if you find something useful not mentioned here and educate other contributors.
- Even though we are going to setup CI, set up the mentioned tools locally to run before you commit.
- If some piece of code is within the scope of your change and it doesn't follow the style guide, refactor it.
- Finally, all this is to ensure code quality without compromising functionality. With that being said, all messages 
  generated by tools are warnings and should be treated as such. It is good practice to address them but dangerous 
  to get too hung up on them.

-----------

### Style Guide

We want to closely follow [Google's Python styleguide](https://google.github.io/styleguide/pyguide.html) 
with few exceptions that will be mentioned here. Although Google's styleguide provides good advice, is quite extensive and could possibly have some rules that might be hard to adhere to in our case. Such rules can be discussed for exemption later and if found troublesome can be changed.

#### Exceptions/Additions to Google's Python styleguide

##### 2.20 Py 2/3 cross compatibility

- Take some effort to ensure code compatibility between Python 2.6.x and 3.x+ . Follow section 2.20,
  use future and six libraries to write cross compatible code. Note, we can drop this compatibility,
  once we finish porting to Py3.
  
  Handy guide: http://python-future.org/compatible_idioms.html 
  
##### 3.2 Line length

Maximum line length is *119 characters*.

##### 3.8 Code documentation

We want to follow Google's doc format as mentioned in section [3.8](https://google.github.io/styleguide/pyguide.html#38-comments-and-docstrings). More examples of such documentation style can be found at

- https://www.sphinx-doc.org/en/master/usage/extensions/example_google.html#example-google 
- https://sphinxcontrib-napoleon.readthedocs.io/en/latest/example_google.html

Adding types to arguments in docstrings can get difficult with complex types. There are some open questions on how
to mention member types for containers like lists, dict, tuple, etc. The sphinx guidelines don't elaborate much on this.
There are many ways we can go

```python
# Consider some function like 
def do_something(texts, attrs, ids=None):
    pass

# Format 1: Specifying member types
texts (list(str or unicode)): list of strings or unicode literals
attrs (tuple(int, str, float) or None):  tuple with three members - int, str, float
ids (dict(str, int) or None, optional): dict mapping str to int. Defaults to None

# Format 2: Not specifying member types and explaining in the explanation part
texts (list): list containing str or unicode literals
attrs (tuple or None): tuple containing
	int: integer
	str: string
	float: float
ids (dict or None, optional): dict with str keys and int values. Defaults to None

# Format 3: Using PEP 484
texts (List[Union[str, unicode]]): list of strings or unicode literals
attrs (Optional[Tuple[int, str, float]]): tuple with three members - int, str, float
ids (Optional[Dict[str, int]], optional): dict mapping str to int. Defaults to None
```

Note that 1st and 3rd are compact and understood by IDEs like PyCharm but harder to read.
2nd format is easier to read but would need more explaning about the types in explanation part.
3rd style is defined by PEP 484, it is much more consistent and already has a detailed guideline
on how to write complex types.

Notice it uses `Optional[]` which indicates this argument can take `None` value;
`Union[]` where members can be multiple types;
whereas 1st format needs the use of `or` to mention multiple types explicitly. 
The `, optional` part indicates if the argument is optional (default) argument for the function 
and is present in all three formats. It is different from `Optional[]`.
Also note that `List, Union, Tuple, Dict` all are in title case. These are defined in `typing` module

It is highly recommended that you go through
[Type Annotation cheat sheet for Python 2](https://mypy.readthedocs.io/en/latest/cheat_sheet.html)
to understand how 3rd format's type annotation works.
We would recommend you follow Format 2 plus adding type annotations with `typing`.

##### 3.10 Strings

Use double quotes `"` for string literals so it becomes easy to copy to other formats (like JSON, C++, Java). It is
okay to use sinngle quote to to avoid the need to `\\` escape within the string

---

### Some other documents to read

- https://docs.python-guide.org/writing/style/ - From the Hitchhiker's Guide to Python highlights
  some good practices that are a subset of the above guideline.
- [Haptik's Python best practices](haptik_python_best_practices.md) - Haptik's internal best practices document,
  again a subset of above guideline, shows some examples of what to do and what not to do.

---

### Tools

We will use [flake8](http://flake8.pycqa.org/en/latest/), [pylint](http://pylint.pycqa.org/en/latest/),
and optionally [mypy](https://mypy.readthedocs.io/en/latest/python2.html) (encouraged but not required)

> Note: Our CI does not run pylint because it is too verbose. However, we would still recommend running it locally

1. **Running flake8**

    flake8 checks for pep8 compatibility

    ```shell
    flake8 --max-line-length=119 <path/to/python file/package/folder>
    ```
    Here is a sample run on `datastore` package

    ```
    flake8 --max-line-length=119 datastore/
    ```

    ```
    datastore/__init__.py:1:1: F401 &#39;datastore.DataStore&#39; imported but unused
    datastore/__init__.py:1:1: F401 'datastore.DataStore' imported but unused
    datastore/exceptions.py:73:1: E302 expected 2 blank lines, found 1
    datastore/exceptions.py:73:1: F811 redefinition of unused 'EngineNotImplementedException' from line 22
    datastore/exceptions.py:84:1: F811 redefinition of unused 'IndexForTransferException' from line 44
    datastore/exceptions.py:94:1: F811 redefinition of unused 'AliasForTransferException' from line 54
    datastore/datastore.py:93:121: E501 line too long (126 > 119 characters)
    datastore/datastore.py:94:121: E501 line too long (131 > 119 characters)
    datastore/elastic_search/__init__.py:1:1: F401 'connect' imported but unused
    datastore/elastic_search/__init__.py:2:1: F401 'create' imported but unused
    datastore/elastic_search/__init__.py:3:1: F401 'populate' imported but unused
    datastore/elastic_search/__init__.py:4:1: F401 'query' imported but unused
    datastore/elastic_search/__init__.py:5:1: F401 'transfer' imported but unused
    datastore/elastic_search/__init__.py:5:16: W292 no newline at end of file
    datastore/elastic_search/populate.py:4:1: F403 'from ..utils import *' used; unable to detect undefined names
    datastore/elastic_search/populate.py:29:21: F405 'get_files_from_directory' may be undefined, or defined from star imports: ..utils
    datastore/elastic_search/populate.py:31:29: F405 'os' may be undefined, or defined from star imports: ..utils
    datastore/elastic_search/populate.py:61:21: F405 'get_files_from_directory' may be undefined, or defined from star imports: ..utils
    datastore/elastic_search/populate.py:63:29: F405 'os' may be undefined, or defined from star imports: ..utils
    datastore/elastic_search/populate.py:90:24: F405 'defaultdict' may be undefined, or defined from star imports: ..utils
    datastore/elastic_search/populate.py:92:22: F405 'read_csv' may be undefined, or defined from star imports: ..utils
    datastore/elastic_search/populate.py:179:22: F405 'os' may be undefined, or defined from star imports: ..utils
    datastore/elastic_search/populate.py:180:22: F405 'os' may be undefined, or defined from star imports: ..utils
    datastore/elastic_search/populate.py:191:50: F405 'remove_duplicate_data' may be undefined, or defined from star imports: ..utils
    datastore/elastic_search/populate.py:193:8: F405 'os' may be undefined, or defined from star imports: ..utils
    datastore/elastic_search/populate.py:193:42: F405 'os' may be undefined, or defined from star imports: ..utils
    datastore/elastic_search/populate.py:194:9: F405 'os' may be undefined, or defined from star imports: ..utils
    datastore/elastic_search/populate.py:263:121: E501 line too long (124 > 119 characters)
    datastore/elastic_search/populate.py:267:121: E501 line too long (127 > 119 characters)
    datastore/elastic_search/populate.py:297:121: E501 line too long (136 > 119 characters)
    ```

    As you see it caught redefinitions so some exceptions that were missed in Code review, `*` import, too long lines,
    `os` import being used from another import. All of this could be easily fixed. 

    **Usually we would want to fix most flake8 reported issues, unless there are some strong reasons to ignore 
    some of them**

2. **Running pylint just for Python 2 and 3 compatibility checks**
 
    Before we run the pylint check for all possible checks, we would want to fix code compatibility issues first as
    more often than not, there are very few issues and once fixed it is unlikely, they pop up again.
    
    To get all compatibility issues, run
    
    ```shell
    pylint --rcfile=~/chatbot_ner/.pylintrc --py3k  -f parseable -r n <path/to/python file/package/folder>
    ```
    
    It is advised to **run it only on changed files**.
    
    **Usually we would want to fix most compatibility reported issues,
      unless there are some strong reasons to ignore some of them**
    
3. **Running pylint**

    pylint checks code for pep8 compatibility plus code complexity and code smell issues.

    ```shell
    pylint --rcfile=~/chatbot_ner/.pylintrc -f parseable -r n <path/to/python file/package/folder>
    ```

    Note you might need to adjust the `—rcfile` path depending on where you are running the command from. 
    Always use the `.pylintrc` provided in the repo. That way we can ensure everyone is using the same config

    Once you start using pylint, you would see that it's **brutal** with your code analysis, 
    so it is advised to **run it only on changed files**. Remember that even if pylint is too picky and strict 
    there is often useful advice for readability and code style in pylint's output. Not all issues need to be strictly 
    resolved but considering them would help. It is recommended to read 
    [pylint's tutorial](http://pylint.pycqa.org/en/latest/tutorial.html).

    Here is an example run on `datastore` package

    ```shell
    cd chatbot_ner/
    pylint --rcfile=.pylintrc -f parseable -r n datastore/
    ```

    ```
    Using config file ~/chatbot_ner/.pylintrc
    ************* Module datastore
    datastore/__init__.py:1: [C0111(missing-docstring), ] Missing module docstring
    datastore/__init__.py:1: [W0403(relative-import), ] Relative import 'datastore', should be 'datastore.datastore'
    ************* Module datastore.constants
    datastore/constants.py:1: [C0111(missing-docstring), ] Missing module docstring
    datastore/constants.py:2: [C0411(wrong-import-order), ] standard import "import os" should be placed before "import elasticsearch"
    ************* Module datastore.utils
    datastore/utils.py:1: [C0111(missing-docstring), ] Missing module docstring
    ************* Module datastore.exceptions
    datastore/exceptions.py:1: [C0111(missing-docstring), ] Missing module docstring
    ...
    datastore/exceptions.py:185: [W0231(super-init-not-called), DeleteIndexFromAliasException.__init__] __init__ method from base class 'Exception' is not called
    ************* Module datastore.datastore
    datastore/datastore.py:341: [W0511(fixme), ] TODO: repopulate code for crf index missing
    datastore/datastore.py:93: [C0301(line-too-long), ] Line too long (126/119)
    datastore/datastore.py:94: [C0301(line-too-long), ] Line too long (131/119)
    datastore/datastore.py:1: [C0111(missing-docstring), ] Missing module docstring
    datastore/datastore.py:3: [W0403(relative-import), ] Relative import 'elastic_search', should be 'datastore.elastic_search'
    datastore/datastore.py:457: [W1201(logging-not-lazy), DataStore.get_crf_data_for_entity_name] Specify string format arguments as logging function parameters
    datastore/datastore.py:475: [W1201(logging-not-lazy), DataStore.get_crf_data_for_entity_name] Specify string format arguments as logging function parameters
    ************* Module datastore.commands.populate_datastore
    datastore/commands/populate_datastore.py:1: [C0111(missing-docstring), ] Missing module docstring
    datastore/commands/populate_datastore.py:8: [C0111(missing-docstring), Command] Missing class docstring
    datastore/commands/populate_datastore.py:51: [E1101(no-member), Command.handle] Instance of 'Style' has no 'ERROR' member
    ************* Module datastore.commands.delete_entity_data_datastore
    datastore/commands/delete_entity_data_datastore.py:1: [C0111(missing-docstring), ] Missing module docstring
    datastore/commands/delete_entity_data_datastore.py:5: [C0111(missing-docstring), Command] Missing class docstring
    datastore/commands/delete_entity_data_datastore.py:22: [E1101(no-member), Command.handle] Instance of 'Style' has no 'ERROR' member
    ************* Module datastore.commands.delete_datastore
    datastore/commands/delete_datastore.py:1: [C0111(missing-docstring), ] Missing module docstring
    datastore/commands/delete_datastore.py:5: [C0111(missing-docstring), Command] Missing class docstring
    ************* Module datastore.commands.repopulate_datastore
    datastore/commands/repopulate_datastore.py:1: [C0111(missing-docstring), ] Missing module docstring
    datastore/commands/repopulate_datastore.py:8: [C0111(missing-docstring), Command] Missing class docstring
    datastore/commands/repopulate_datastore.py:53: [E1101(no-member), Command.handle] Instance of 'Style' has no 'ERROR' member
    ************* Module datastore.commands.create_datastore
    datastore/commands/create_datastore.py:1: [C0111(missing-docstring), ] Missing module docstring
    datastore/commands/create_datastore.py:5: [C0111(missing-docstring), Command] Missing class docstring
    ************* Module datastore.elastic_search.create
    datastore/elastic_search/create.py:1: [C0111(missing-docstring), ] Missing module docstring
    datastore/elastic_search/create.py:1: [W0403(relative-import), ] Relative import 'utils', should be 'datastore.elastic_search.utils'
    datastore/elastic_search/create.py:26: [W0703(broad-except), delete_index] Catching too general exception Exception
    datastore/elastic_search/create.py:91: [W0703(broad-except), _create_index] Catching too general exception Exception
    ************* Module datastore.elastic_search.query
    datastore/elastic_search/query.py:286: [W1401(anomalous-backslash-in-string), ] Anomalous backslash in string: '\s'. String constant might be missing an r prefix.
    datastore/elastic_search/query.py:1: [C0111(missing-docstring), ] Missing module docstring
    datastore/elastic_search/query.py:50: [R0913(too-many-arguments), full_text_query] Too many arguments (7/5)
    datastore/elastic_search/query.py:2: [C0411(wrong-import-order), ] standard import "import re" should be placed before "from six import string_types"
    datastore/elastic_search/query.py:3: [C0411(wrong-import-order), ] standard import "import collections" should be placed before "from six import string_types"
    datastore/elastic_search/query.py:6: [C0411(wrong-import-order), ] first party import "from external_api.constants import SENTENCE_LIST, ENTITY_LIST" should be placed before "from ..constants import ELASTICSEARCH_SEARCH_SIZE, ELASTICSEARCH_VERSION_MAJOR, ELASTICSEARCH_VERSION_MINOR"
    datastore/elastic_search/query.py:7: [C0411(wrong-import-order), ] first party import "from language_utilities.constant import ENGLISH_LANG" should be placed before "from ..constants import ELASTICSEARCH_SEARCH_SIZE, ELASTICSEARCH_VERSION_MAJOR, ELASTICSEARCH_VERSION_MINOR"
    ************* Module datastore.elastic_search.connect
    datastore/elastic_search/connect.py:1: [C0111(missing-docstring), ] Missing module docstring
    ************* Module datastore.elastic_search.__init__
    datastore/elastic_search/__init__.py:5: [C0304(missing-final-newline), ] Final newline missing
    ************* Module datastore.elastic_search
    datastore/elastic_search/__init__.py:1: [C0111(missing-docstring), ] Missing module docstring
    ...
    datastore/elastic_search/__init__.py:5: [W0403(relative-import), ] Relative import 'transfer', should be 'datastore.elastic_search.transfer'
    ************* Module datastore.elastic_search.transfer
    datastore/elastic_search/transfer.py:276: [W0511(fixme), ] TODO - this works differently from other connects, picks scheme from the full URL
    datastore/elastic_search/transfer.py:207: [C0325(superfluous-parens), ] Unnecessary parens after 'while' keyword
    datastore/elastic_search/transfer.py:1: [C0111(missing-docstring), ] Missing module docstring
    datastore/elastic_search/transfer.py:192: [W1201(logging-not-lazy), ESTransfer.check_if_index_exits] Specify string format arguments as logging function parameters
    datastore/elastic_search/transfer.py:247: [R0201(no-self-use), ESTransfer._generate_update_json] Method could be a function
    datastore/elastic_search/transfer.py:464: [W0102(dangerous-default-value), ESTransfer.transfer_specific_entities] Dangerous default value [] as argument
    datastore/elastic_search/transfer.py:471: [W1201(logging-not-lazy), ESTransfer.transfer_specific_entities] Specify string format arguments as logging function parameters
    datastore/elastic_search/transfer.py:477: [W1201(logging-not-lazy), ESTransfer.transfer_specific_entities] Specify string format arguments as logging function parameters
    datastore/elastic_search/transfer.py:480: [W1201(logging-not-lazy), ESTransfer.transfer_specific_entities] Specify string format arguments as logging function parameters
    datastore/elastic_search/transfer.py:484: [W1201(logging-not-lazy), ESTransfer.transfer_specific_entities] Specify string format arguments as logging function parameters
    datastore/elastic_search/transfer.py:492: [W1201(logging-not-lazy), ESTransfer.transfer_specific_entities] Specify string format arguments as logging function parameters
    datastore/elastic_search/transfer.py:2: [C0411(wrong-import-order), ] standard import "import json" should be placed before "import requests"
    ************* Module datastore.elastic_search.utils
    datastore/elastic_search/utils.py:1: [C0111(missing-docstring), ] Missing module docstring
    ************* Module datastore.elastic_search.populate
    datastore/elastic_search/populate.py:114: [C0301(line-too-long), ] Line too long (120/119)
    datastore/elastic_search/populate.py:154: [C0330(bad-continuation), ] Wrong continued indentation (remove 1 space).
                          }
                         |^
    datastore/elastic_search/populate.py:220: [C0330(bad-continuation), ] Wrong continued indentation (remove 1 space).
                               }
                              |^
    datastore/elastic_search/populate.py:263: [C0301(line-too-long), ] Line too long (124/119)
    datastore/elastic_search/populate.py:267: [C0301(line-too-long), ] Line too long (127/119)
    datastore/elastic_search/populate.py:297: [C0301(line-too-long), ] Line too long (136/119)
    datastore/elastic_search/populate.py:332: [C0330(bad-continuation), ] Wrong continued indentation (remove 1 space).
                          }
                         |^
    datastore/elastic_search/populate.py:1: [C0111(missing-docstring), ] Missing module docstring
    datastore/elastic_search/populate.py:4: [W0401(wildcard-import), ] Wildcard import utils
    datastore/elastic_search/populate.py:10: [R0913(too-many-arguments), create_all_dictionary_data] Too many arguments (6/5)
    datastore/elastic_search/populate.py:42: [R0913(too-many-arguments), recreate_all_dictionary_data] Too many arguments (6/5)
    datastore/elastic_search/populate.py:105: [W0703(broad-except), get_variants_dictionary_value_from_key] Catching too general exception Exception
    datastore/elastic_search/populate.py:101: [W0703(broad-except), get_variants_dictionary_value_from_key] Catching too general exception Exception
    datastore/elastic_search/populate.py:74: [W0613(unused-argument), get_variants_dictionary_value_from_key] Unused argument 'kwargs'
    ...
    datastore/elastic_search/populate.py:4: [W0614(unused-wildcard-import), ] Unused import csv from wildcard import
    datastore/elastic_search/populate.py:5: [C0411(wrong-import-order), ] first party import "from language_utilities.constant import ENGLISH_LANG" should be placed before "from ..constants import ELASTICSEARCH_BULK_HELPER_MESSAGE_SIZE, ELASTICSEARCH_SEARCH_SIZE"
    datastore/elastic_search/populate.py:1: [R0801(duplicate-code), ] Similar lines in 2 files
    ==datastore.commands.populate_datastore:11
    ==datastore.commands.repopulate_datastore:13
    ...
    Your code has been rated at 8.30/10
    ```

    As you see it warns about missing documentation, non standard naming conventions, indentation mistakes, long lines,
    import errors, too many arguments, catching general exception etc. The output is quite verbose 
    and not always perfect. For example - we use dynamic imports in some places which pylint can't understand. 
    In such cases it makes sense to ignore such messages. But some errors like `*` imports, indentation mistakes, 
    long lines, too many arguments, relative imports provide good signals to refactor those parts.

    **We do not aim to achieve a 10/10 score from pylint**

> You can use tools like [yapf](https://github.com/google/yapf) or [autopep8](https://github.com/hhatto/autopep8) or 
  Pycharm plugins to fix lint errors or fix them manually. 
  In the end all that matters is fixing them.

4. **Running mypy**

   MyPy helps checking type annotations in code according to [PEP 484](https://www.python.org/dev/peps/pep-0484/). 
   Since we are still using Py 2.x, we cannot use mypy type annotations as we can in Python 3.5+. 
   Only option to annotate is to do so via inline comments. 
   MyPy docs do a good job of explaining how to install, annotate and use it. 
   However, note that running mypy needs Python 3. 
   Adding type annotations is encouraged however we are not enforcing it. 
   We would highly encourage you to read mypy docs. Some handy pages:

   - [Introduction](https://mypy.readthedocs.io/en/latest/introduction.html)
   - [Use with existing codebase](https://mypy.readthedocs.io/en/latest/existing_code.html)
   - [Type Checking for Python2](https://mypy.readthedocs.io/en/latest/python2.html)
   - [Type Annotation cheat sheet for Python 2](https://mypy.readthedocs.io/en/latest/cheat_sheet.html)

   Running mypy

   ```shell
   mypy <path/to/python file/package/folder>
   ```

   ​